from pandas import DataFrame


class StoreDataFrame:
    df = DataFrame(
        {
            "id": [1, 2],
            "name": ["John", "Jane"],
            "description": [
                "John loves writing pytests in vscode",
                "Dit is een Nederlandse tekst",
            ],
        }
    )
    path: str = None

    def store_dataframe(self, base_path: str) -> str:
        pass


class CSVStoreDataFrame(StoreDataFrame):
    def store_dataframe(self, base_path: str) -> str:
        path = f"{base_path}/example.csv"
        self.df.to_csv(path, index=False)

        return path


class JSONStoreDataFrame(StoreDataFrame):
    def store_dataframe(self, base_path: str) -> str:
        path = f"{base_path}/example.json"
        self.df.to_json(path, orient="records", lines=True)

        return path


class PickleStoreDataFrame(StoreDataFrame):
    def store_dataframe(self, base_path: str) -> str:
        path = f"{base_path}/example.pkl"
        self.df.to_pickle(path)

        return path



from typing import Any

from dagster import InputDefinition, Nothing, OutputDefinition
from qdk.base import BaseComponent


class BaseLoaderComponent(BaseComponent):
    compute_function = "load"
    tags = {
        "kind": "loader",
    }
    input_defs = [
        InputDefinition("after", Nothing),
    ]
    output_defs = [
        OutputDefinition(Any, "out"),
    ]

    @classmethod
    def load(cls) -> Any:
        """
        Overwrite this function to load the required object.
        """
        raise NotImplementedError(
            'Make sure you added a "load" function to the component'
        )



from qdk import DataFrameLoader, DataFrameType
from ..utils.dataframe import (
    CSVStoreDataFrame,
    JSONStoreDataFrame,
    PickleStoreDataFrame,
)
from dagster import SolidDefinition, DagsterType


def test_csv_loading(tmpdir):
    """Test loading a csv file.

    Args:
        tmpdir (str): The path to the temporary directory created by pytest
    """
    df_storer = CSVStoreDataFrame()
    df_path = df_storer.store_dataframe(tmpdir)

    loaded_df = DataFrameLoader.load(df_path)
    loaded_ddf = DataFrameLoader.load(df_path, use_dask=True).compute()

    assert loaded_df.equals(df_storer.df)
    assert loaded_ddf.equals(df_storer.df)


def test_json_loading(tmpdir):
    """Test loading a json file.

    Args:
        tmpdir (str): The path to the temporary directory created by pytest
    """
    df_storer = JSONStoreDataFrame()
    df_path = df_storer.store_dataframe(tmpdir)

    loaded_df = DataFrameLoader.load(df_path)
    loaded_ddf = DataFrameLoader.load(df_path, use_dask=True).compute()

    assert loaded_df.equals(df_storer.df)
    assert loaded_ddf.equals(df_storer.df)


def test_pickle_loading(tmpdir):
    """Test loading a pickle file.

    Args:
        tmpdir (str): The path to the temporary directory created by pytest
    """
    df_storer = PickleStoreDataFrame()
    df_path = df_storer.store_dataframe(tmpdir)

    loaded_df = DataFrameLoader.load(df_path)
    loaded_ddf = DataFrameLoader.load(df_path, use_dask=True, repartitions=1).compute()

    assert loaded_df.equals(df_storer.df)
    assert loaded_ddf.equals(df_storer.df)


def test_solid_creation():
    df_loader = DataFrameLoader
    solid = df_loader.to_solid("df_loader")

    assert type(solid) == SolidDefinition


def test_dagster_output_def():
    df_loader = DataFrameLoader
    solid = df_loader.to_solid("df_loader")

    assert len(solid.output_defs) == 1
    assert solid.output_defs[0].dagster_type == DataFrameType



from typing import List, Set, Union

import dask.dataframe as dd
import nltk
import pandas as pd
import spacy
from dagster import Field
from qdk.transform.base import TransformComponent
from qdk.utils.text import clean_text


class NounPhraseComponent(TransformComponent):
    replace_pos = {
        "ADJ": "JJ",
        "NOUN": "NN",
        # "PROPN": "NN",
    }
    noun_phrase_regex = "NP: {<NN.*|JJ>*<NN.*>}"
    config_schema = {
        "text_columns": Field(
            list,
            default_value=["text"],
            description="Columns to tokenize",
        ),
        "spacy_model": Field(
            str,
            default_value="nl_core_news_lg",
            description="The name of the spacy language model (nl_core_news_lg)",
        ),
    }

    @classmethod
    def extract_noun_phrases(cls, text: str, nlp: spacy.Language) -> List[str]:
        """This function takes a string as input and outputs a list of
        noun phrase candidates:

        "Dit is een zin met tekst" -> {"zin", "tekst"}

        Args:
            text (str): The text you would like to transform.
            nlp (spacy.Language): A spacy language model. Used for POS tagging.

        Returns:
            List[str]: The noun phrases found in the text.
        """
        # Clean the text before extracting phrases
        cleaned_text = clean_text(text=text)

        # Use spacy to parse and tokenize the text
        # Possibly replace pos tags
        doc = nlp(cleaned_text)
        doc = [
            (token.text.lower(), cls.replace_pos.get(token.pos_, token.pos_))
            for token in doc
        ]

        # Use the nltk regex parser to find all noun phrases
        parser = nltk.RegexpParser(cls.noun_phrase_regex)
        trees = parser.parse_sents([doc])

        # Storage for all key phrase candidates
        noun_phrases = []

        for tree in trees:
            for subtree in tree.subtrees(
                filter=lambda t: t.label() == "NP"
            ):  # For each nounphrase
                # Concatenate the token with a space
                noun_phrases.append(" ".join(word for word, _ in subtree.leaves()))

        # Filter out keyphrases that have more than 5 tokens
        noun_phrases = [
            noun_phrase for noun_phrase in noun_phrases if len(noun_phrase.split()) <= 5
        ]

        return noun_phrases

    @classmethod
    def transform(
        cls,
        df: Union[pd.DataFrame, dd.DataFrame],
        text_columns: List[str] = ["text"],
        spacy_model: str = "nl_core_news_lg",
    ) -> Union[pd.DataFrame, dd.DataFrame]:
        nlp = spacy.load(
            spacy_model,
            disable=[
                "parser",
                "ner",
                "lemmatizer",
                "textcat",
            ],
        )

        # Use dask to parallelize the keyword extraction over the workers
        if type(df) == dd.DataFrame:
            for text_column in text_columns:
                df[f"_nouns_{text_column}"] = df.map_partitions(
                    lambda _df: _df[text_column].apply(
                        cls.extract_noun_phrases, args=(nlp,)
                    )
                )
        # Pandas apply function
        else:
            for text_column in text_columns:
                df[f"_nouns_{text_column}"] = df[text_column].apply(
                    cls.extract_noun_phrases, args=(nlp,)
                )

        # Return the dataframe with the keywords
        return df



from abc import ABC, abstractstaticmethod
from typing import Any, Dict, List, Optional

from dagster import AssetMaterialization, EventMetadata
from dask.dataframe import DataFrame as DaskDataFrame
from pandas import DataFrame

from .mlflow import MLFlowRun
from .utils.typing import is_dask_dataframe, is_mlflow_run, is_pandas_dataframe


def list_to_string(list: List[Any]) -> str:
    return f"[{', '.join(list)}]"


class MetaData(ABC):
    @abstractstaticmethod
    def metadata(object: Any) -> Dict[str, Any]:
        pass


class PandasDataFrameMetaData(MetaData):
    @staticmethod
    def create_metadata(object: DataFrame) -> Dict[str, Any]:
        if not is_pandas_dataframe(object):
            raise ValueError("Object is not a pandas dataframe")

        return {
            "columns": list_to_string(object.columns),
            "dask": 0,
            "length": len(object),
            "head": EventMetadata.md(object.head().to_markdown()),
            "tail": EventMetadata.md(object.tail().to_markdown()),
        }


class DaskDataFrameMetaData(MetaData):
    @staticmethod
    def create_metadata(object: DaskDataFrame) -> Dict[str, Any]:
        if not is_dask_dataframe(object):
            raise ValueError("Object is not a dask dataframe")

        return {
            "columns": list_to_string(object.columns), 
            "dask": 1,
        }


class MLFlowRunMetaData(MetaData):
    @staticmethod
    def create_metadata(object: MLFlowRun) -> Dict[str, Any]:
        if not is_mlflow_run(object):
            raise ValueError("Object is not a mlflow run")

        return {
            "run_id": object.run_id,
            "model_flavors": list_to_string(object.model_flavors),
            "run_url": EventMetadata.url(object.run_url),
            "run_url_local": EventMetadata.url(object.run_url_local),
            "input_schema": EventMetadata.md(object.input_schema.to_markdown()),
            "output_schema": EventMetadata.md(object.output_schema.to_markdown()),
        }


class Materializer:
    def __init__(self, asset_key: str, object: Any):
        self.asset_key = asset_key
        self.object = object

        if is_pandas_dataframe(object):
            self.metadata = PandasDataFrameMetaData.create_metadata(object)
        elif is_dask_dataframe(object):
            self.metadata = DaskDataFrameMetaData.create_metadata(object)
        elif is_mlflow_run(object):
            self.metadata = MLFlowRunMetaData.create_metadata(object)
        else:
            self.metadata = {}

    def materialize(self) -> Optional[AssetMaterialization]:
        if self.metadata:
            return AssetMaterialization(
                asset_key=self.asset_key,
                metadata=self.metadata,
            )



import pandas as pd
import dask.dataframe as dd
from dagster import Field, OutputDefinition, Permissive
from qdk.dagster_types import MLFlowRunType
from qdk.training.base import TrainingComponent
from qdk.mlflow import MLFlowRun
from typing import Any, Dict, Union
from sklearn.base import BaseEstimator
from mlflow import set_experiment, autolog, start_run, xgboost, sklearn


class MLFlowTrainingComponent(TrainingComponent):
    # required_resource_keys = {"mlflow"}
    output_defs = [OutputDefinition(MLFlowRunType, "mlflow_run")]
    config_schema = {
        "experiment_name": Field(
            str,
            default_value="default",
            description="The mlflow experiment name, if it doesn't exists it will be created.",
        )
    }

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        y: Union[pd.Series, dd.Series],
        model: BaseEstimator,
        experiment_name: str = "default",
    ):
        set_experiment(experiment_name)
        autolog(disable=False)

        with start_run() as run:
            model.fit(X, y)

        autolog(disable=True)
        return MLFlowRun(run)



from dask.array.core import Array as DaskArray
from dask.dataframe import DataFrame as DaskDataFrame
from dask.dataframe import Series as DaskSeries
from gensim.models.word2vec import Word2Vec
from numpy import ndarray
from pandas import DataFrame, Series
from qdk.mlflow import MLFlowRun
from sklearn.base import BaseEstimator


def is_dataframe(value) -> bool:
    """Function to check whether the incoming value is a pandas or dask dataframe.

    Args:
        value (Any): The incoming dagster value.

    Returns:
        bool: Whether the input is a dataframe or not.
    """
    return (
        isinstance(value, DataFrame)
        or isinstance(value, DaskDataFrame)
        or isinstance(value, DaskArray)
        or isinstance(value, ndarray)
    )


def is_series(value) -> bool:
    """Function to check whether the incoming value is a pandas or dask series.

    Args:
        value (Any): The incoming dagster value.

    Returns:
        bool: Whether the input is a series or not.
    """
    return (
        isinstance(value, Series)
        or isinstance(value, DaskSeries)
        or isinstance(value, DaskArray)
        or isinstance(value, ndarray)
    )


def is_pandas_dataframe(value):
    return isinstance(value, DataFrame)


def is_dask_dataframe(value):
    return isinstance(value, DaskDataFrame)


def is_model(value):
    return (
        isinstance(value, BaseEstimator)
        or isinstance(value, Word2Vec)
        or isinstance(value, MLFlowRun)
    )


def is_mlflow_run(value):
    return isinstance(value, MLFlowRun)



from dagster import (
    DagsterType,
    ModeDefinition,
    PipelineExecutionResult,
    SolidDefinition,
    execute_pipeline,
    pipeline,
)
from dagster.core.execution.results import SolidExecutionResult
from dask.dataframe import DataFrame as DaskDataFrame
from pandas import DataFrame
from qdk import DataFrameLoader, YakeComponent, qdk_io_manager

from ..utils.dataframe import CSVStoreDataFrame

loader_solid = DataFrameLoader.to_solid(name="loader")

keyword_solid = YakeComponent.to_solid(
    name="keyword",
)


@pipeline(mode_defs=[ModeDefinition(resource_defs={"io_manager": qdk_io_manager})])
def nlp_pipeline():
    df = loader_solid()
    keyword = keyword_solid(df)


def run_pipeline(tmpdir: str, use_dask: bool) -> SolidExecutionResult:
    # Create a temporary dataframe and store it in the temp directory
    df_storer = CSVStoreDataFrame()
    df_path = df_storer.store_dataframe(tmpdir)

    # Execute the pipeline
    result = execute_pipeline(
        nlp_pipeline,
        run_config={
            "solids": {
                "loader": {
                    "config": {
                        "uri": df_path,
                        "use_dask": use_dask,
                    }
                },
                "keyword": {
                    "config": {"language": "en", "text_columns": ["description"]}
                },
            }
        },
    )

    return result


def test_nlp_pipeline(tmpdir):
    # Execute the pipeline
    result = run_pipeline(
        tmpdir,
        use_dask=False,
    )

    result_df = result.output_for_solid("keyword", "df")

    assert result.success
    assert isinstance(result, PipelineExecutionResult)
    assert isinstance(result_df, DataFrame)


def test_nlp_dask_pipeline(tmpdir):
    # Execute the pipeline
    result = run_pipeline(
        tmpdir,
        use_dask=True,
    )

    result_df = result.output_for_solid("keyword", "df")

    assert result.success
    assert isinstance(result, PipelineExecutionResult)
    assert isinstance(result_df, DaskDataFrame)



from typing import Optional, Tuple, Union

import dask.dataframe as dd
import pandas as pd
from dagster import InputDefinition, Nothing
from dagster import Optional as DagsterOptional
from dagster import OutputDefinition
from qdk.base import BaseComponent
from qdk.dagster_types import DataFrameType, ModelType, SeriesType

from sklearn.base import BaseEstimator


class TrainingComponent(BaseComponent):
    compute_function = "train"
    tags = {
        "kind": "training",
    }
    input_defs = [
        InputDefinition(
            "X",
            DataFrameType,
        ),
        InputDefinition(
            "y",
            DagsterOptional[SeriesType],
        ),
        InputDefinition(
            "model",
            DagsterOptional[ModelType],
        ),
    ]
    output_defs = [
        OutputDefinition(
            ModelType,
            "model",
        ),
        OutputDefinition(
            DataFrameType,
            "data",
            is_required=False,
        ),
    ]

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        model: Optional[BaseEstimator] = None,
        y: Optional[Union[pd.Series, dd.Series]] = None,
    ) -> Tuple[BaseEstimator, Optional[Union[pd.DataFrame, dd.DataFrame]]]:
        raise NotImplementedError(
            'Make sure you added a "train" function to the component'
        )



from qdk.inference.base import InferenceComponent
from qdk.dagster_types import MLFlowRunType, DataFrameType, SeriesType
from qdk.mlflow import MLFlowRun
from dagster import InputDefinition, OutputDefinition
from pandas import DataFrame, Series
from dask.dataframe import DataFrame as DaskDataFrame, Series as DaskSeries
from typing import Union


class MLFlowInferenceComponent(InferenceComponent):
    compute_function = "predict"
    input_defs = [
        InputDefinition("X", DataFrameType),
        InputDefinition("mlflow_run", MLFlowRunType),
        InputDefinition("y_true", SeriesType, default_value=None),
    ]
    output_defs = [OutputDefinition(DataFrameType, "X_hat")]

    @classmethod
    def predict(
        cls,
        X: Union[DataFrame, DaskDataFrame],
        mlflow_run: MLFlowRun,
        y_true: Union[Series, DaskSeries] = None,
    ) -> Union[DataFrame, DaskDataFrame]:
        # Use the mlflow model to predict
        X_hat: Union[Series, DaskSeries] = mlflow_run.model.predict(X)

        X["y_hat"] = X_hat

        # If y_true is provided, add it to the output
        if y_true is not None:
            X["y_true"] = y_true

        return X



from typing import Union

import dask.dataframe as dd
import pandas as pd
from dagster import InputDefinition, OutputDefinition
from qdk.base import BaseComponent
from qdk.dagster_types import DataFrameType


class TransformComponent(BaseComponent):
    compute_function = "transform"
    tags = {
        "kind": "transform",
    }
    input_defs = [
        InputDefinition("df", DataFrameType),
    ]
    output_defs = [
        OutputDefinition(DataFrameType, "df"),
    ]

    @classmethod
    def transform(
        cls,
        df: Union[pd.DataFrame, dd.DataFrame],
        **config,
    ) -> Union[pd.DataFrame, dd.DataFrame]:
        raise NotImplementedError(
            'Make sure you added a "transform" function to the component'
        )
